/*
 *
 *
 */
#include "common.h"
#include "usb.h"
#include "usbhw.h"

/*
 * internal buffer size for endpoint 0
 *   must be greater than max_packet_size described in device desctiptor
 */
#define MAX_PACKET_SIZE_EP0    64

/*
 * Max Number of Interfaces
 */
#define USB_N_IF          4

/*
 * select default power setting
 *   0: Bus-powered
 *   1: Self-powered
 */
#define USB_POWER         0

/*
 * make state machine for CONTROL-PIPE
 */
#if 0
#define SUDS_IDLE     0		/* only SETUP will be accepted */
#define SUDS_DATAIN   1		/* expect IN from host */
#define SUDS_DATAOUT  2		/* expect OUT from host */
#define SUDS_STATOUT  3		/* expect OUT from host */
#define SUDS_STATIN   4		/* expect IN from host */
[SUDS_IDLE]
 | SETUP(OUT)
 |     WITH DATA
				[SUDS_DATAOUT]
 |         IN->stall
 |         OUT
 |             ->stall if buffer over
 |             ->recv data until specified length--> ep0_data_ready=TRUE
						[SUDS_STATIN]
 |                 IN->send status
 |                 OUT->stall
 |     WITHOUT DATA
				[SUDS_STATIN] <========
 |         IN->send status
 |         OUT->stall
 | SETUP(IN)
 |     WITHOUT DATA
				[ucore_STALL_EP0()]
 |         IN->stall
 |         OUT->stall
 |     WITH DATA
				[SUDS_DATAIN] <=========
 |         OUT->stall
 |         IN->send data (zero-length-packet if needed)
										(datasize<requestsize) &&
										(datasize % packetsize == 0)
					[SUDS_STATOUT]
 |             OUT->send status
 |             IN->stall
SetupPacket.dir = (SETUPDAT[0] >> 7) & 0x01;
EP0BUF_count = SETUPDAT[7,6];
 parse SETUP request==>setup data phase==> uhw_ReadEP_finish()
       IN   send data from pointer, enter status
       OUT  recv data from pointer==>parse the data==>swhich status
IN0BUF[],OUT0BUF[]
#endif /* 0 */

///*
// * structure for USB endpoint
// */
//typedef struct _USB_EP_DATA {
//	u_char  *pData;
//	u_short Count;
//} USB_EP_DATA;

/*
 *
 */
#define EP_CTRL_IN   (0x80 | 0)
#define EP_CTRL_OUT  (0x00 | 0)

#define finish_to_parse_setup_request()   uhw_ReadEP_finish(EP_CTRL_OUT)

/*
 * functions
 */
void ucore_InitialCore(void);
void ucore_ResetCore(void);
void ucore_WakeUp_host(void);
void ucore_isr_Ep0(u_long event);
static void ucore_DataInStage(void);
static void ucore_DataOutStage(void);
static void ucore_StatusInStage(void);
static void ucore_StatusOutStage(void);
void ucore_feed_SetupCommand(void);
void ucore_SetupCommand(void);
static /*__inline*/ int ucore_GetStatus(void);
static /*__inline*/ int ucore_SetClrFeature(u_long sc);
static /*__inline*/ int ucore_GetDescriptor(void);
static u_char *ucore_GetStringDscr(u_char* top_string_desc, u_char index);
static /*__inline*/ int ucore_SetConfiguration(void);
static /*__inline*/ int ucore_SetInterface(void);
void ucore_STALL_EP0(void);

/*
 * variables
 */
u_char *USB_DeviceDescriptor;
u_char *USB_ConfigDescriptor;
u_char *USB_StringDescriptor;
u_long max_packet_size_ep0;
           u_short USB_DeviceStatus;
/*static*/ u_char  USB_DeviceAddress;
           u_char  USB_Configuration;
/*static*/ u_long  USB_EndPointMask;
/*static*/ u_long  USB_EndPointHalt;
/*static*/ u_char  USB_NumInterfaces;
/*static*/ u_char  USB_AltSetting[USB_N_IF];

/*static*/ u_char EP0BUF[MAX_PACKET_SIZE_EP0];
/*static*/ u_char *EP0BUF_ptr;
/*static*/ u_long EP0BUF_count;
/*static*/ u_long EP0BUF_len_requested;
/*static*/ int    EP0BUF_add_zero_packet;
///*static*/ USB_EP_DATA      EP0Data;
u_char SETUPDAT[8];
USB_SETUP_PACKET SetupPacket;
int GotSUD;

/*
 * externs
 */
/* in dscr.S */
extern u_long *DeviceDscr, *ConfigDscr, *StringDscr;
/* in usbhw.c */
extern void uhw_WakeUp(void);
extern void uhw_SetAddress(u_char adr);
extern void uhw_Configure(int cfg);
extern void uhw_ConfigEP(u_char *pEPD);
extern void uhw_EnableEP(u_char EPNum);
extern void uhw_DisableEP(u_char EPNum);
extern void uhw_ResetEP(u_char EPNum);
extern void uhw_SetStallEP(u_char EPNum);
extern void uhw_ClrStallEP(u_char EPNum);
extern u_long uhw_WriteEP(u_char EPNum, u_char *pData, u_long cnt);
extern void uhw_arm_WriteEP(u_char EPNum);
extern u_long uhw_ReadEP(u_char EPNum, u_char *pData);
extern void uhw_ReadEP_finish(u_char EPNum);
/* in vcomuser.c */
extern void vcom_USB_Configure_Event(void);
extern void vcom_USB_Interface_Event(void);
extern void vcom_USB_Feature_Event(void);
extern int vcom_SetSIOSetup(u_char Cmd, u_char Data);


/*
 * initial USB Core
 */
void ucore_InitialCore(void)
{
	USB_DeviceDescriptor = (u_char *)&DeviceDscr;
	USB_ConfigDescriptor = (u_char *)&ConfigDscr;
	USB_StringDescriptor = (u_char *)&StringDscr;

	max_packet_size_ep0 = USB_DeviceDescriptor[7];
}

/*
 *  Reset USB Core
 *    Parameters:      None
 *    Return Value:    None
 */
void ucore_ResetCore(void)
{
	USB_DeviceStatus  = USB_POWER;
	USB_DeviceAddress = 0;
	USB_Configuration = 0;

	USB_EndPointMask  = 0x00010001;
	USB_EndPointHalt  = 0x00000000;
}

/*
 *  USB Remote Wakeup Function
 *    Return Value:    None
 */
void ucore_WakeUp_host(void)
{
	if(USB_DeviceStatus & USB_GETSTATUS_REMOTE_WAKEUP) {
		uhw_WakeUp();
	}
}

/*
 *  USB Endpoint 0 Event Callback
 *    Parameter:       event
 */
void ucore_isr_Ep0(u_long event)
{
	switch(event) {
	case USB_EVT_SETUP:
rprintf("EVT_SETUP\r\n");
			GotSUD = TRUE;
			break;

	case USB_EVT_OUT:
rprintf("EVT_OUT\r\n");
			if(SetupPacket.dir == REQUEST_HOST_TO_DEVICE) {
				if(EP0BUF_count) {
	// zero-length-packet
					ucore_DataOutStage();
					if(EP0BUF_count == 0) {
//						switch(*((u_short *)&SetupPacket)) {
//						default:
//								goto stall_i;
//						}
						/*
						 * make status-in ready
						 */
						ucore_StatusInStage();
					}
				}
			} else {
				/*
				 * always ignore 'status-OUT-has-come'
				 *   (following next SETUP shall be always
				 *    'Packet over-written', don't mind)
				 */
				//ucore_StatusOutStage();
			}
			break;

	case USB_EVT_IN:
rprintf("EVT_IN\r\n");
			if(SetupPacket.dir == REQUEST_DEVICE_TO_HOST) {
				if((EP0BUF_count != 0) || EP0BUF_add_zero_packet) {
					ucore_DataInStage();
				}
			} else {
				/*
				 * 'status-in-finished' here, nothing to do.
				 * only about SET_ADDR request, update device address here.
				 */
				if(USB_DeviceAddress & 0x80) {
					USB_DeviceAddress &= 0x7F;
					uhw_SetAddress(USB_DeviceAddress);
				}
			}
			break;

	case USB_EVT_IN_STALL:
			uhw_ClrStallEP(EP_CTRL_IN);
			break;

	case USB_EVT_OUT_STALL:
			uhw_ClrStallEP(EP_CTRL_OUT);
			break;

//	default:
//stall_i:
//			uhw_SetStallEP(EP_CTRL_IN);
//			EP0BUF_count = 0;
//			break;
	}
}

/*
 *  USB Request - Data In Stage
 *    Parameters:      None (global EP0Data)
 *    Return Value:    None
 */
static void ucore_DataInStage(void)
{
	u_long cnt;

	/*
	 * calc data length for this packet
	 */
	if(EP0BUF_count > max_packet_size_ep0) {
		cnt = max_packet_size_ep0;
	} else {
		cnt = EP0BUF_count;
	}
rprintf("DataIn(): [n=%d] %02x %02x..\r\n", cnt, EP0BUF_ptr[0], EP0BUF_ptr[1]);

	/*
	 * write the data into the buffer
	 */
	uhw_WriteEP(EP_CTRL_IN, EP0BUF_ptr, cnt);

	/*
	 *
	 */
	if(EP0BUF_add_zero_packet) {
		/*
		 * reset this flag
		 */
		EP0BUF_add_zero_packet = FALSE;
	} else {
		/*
		 * advance pointers for next operation
		 */
		EP0BUF_ptr += cnt;
		EP0BUF_count -= cnt;
		EP0BUF_len_requested -= cnt;

		/*
		 * add zero-length-packet if the data length is just fit
		 * to packet size while the host expect more.
		 */
		if(
			(cnt == max_packet_size_ep0) &&
			(EP0BUF_count == 0) &&
			(EP0BUF_len_requested != 0)
		) {
			EP0BUF_add_zero_packet = TRUE;
		}
	}

	/*
	 * at last, arm the endpoint
	 *   (to prevent 'transfer-finished' interrupt among above process)
	 */
	uhw_arm_WriteEP(EP_CTRL_IN);
}

/*
 *  USB Request - Data Out Stage
 *    Parameters:      None (global EP0Data)
 *    Return Value:    None
 */
// finish EP0 before call this
static void ucore_DataOutStage(void)
{
	u_long cnt;

rprintf("DataOutStage():");
	cnt = uhw_ReadEP(EP_CTRL_OUT, EP0BUF_ptr);
	uhw_ReadEP_finish(EP_CTRL_OUT);
rprintf(" %02x %02x..", EP0BUF_ptr[0], EP0BUF_ptr[1]);
rprintf(" %d bytes\r\n", cnt);
	EP0BUF_ptr += cnt;
	EP0BUF_count -= cnt;
}

/*
 *  USB Request - Status In Stage
 *    Parameters:      None
 *    Return Value:    None
 */
static void ucore_StatusInStage(void)
{
rprintf("StatusInStage()\r\n");
	uhw_WriteEP(EP_CTRL_IN, NULL, 0);
	uhw_arm_WriteEP(EP_CTRL_IN);
}

/*
 *  USB Request - Status Out Stage
 *    Parameters:      None
 *    Return Value:    None
 */
static void ucore_StatusOutStage(void)
{
rprintf("StatusOutStage()\r\n");
	uhw_ReadEP(EP_CTRL_OUT, EP0BUF);
	uhw_ReadEP_finish(EP_CTRL_OUT);
}

/*
 * Device request parser
 */
void ucore_feed_SetupCommand(void)
{
	uhw_ReadEP(EP_CTRL_OUT, SETUPDAT);
//	uhw_ReadEP_finish(EP_CTRL_OUT);
// try, how act does he ? ==> status OUT is pend !
	GotSUD = FALSE;

	SetupPacket.dir   = (SETUPDAT[0] >> 7) & 0x01;
	SetupPacket.type  = (SETUPDAT[0] >> 5) & 0x03;
	SetupPacket.recip = (SETUPDAT[0] >> 0) & 0x1f;
	SetupPacket.bRequest = SETUPDAT[1];
	SetupPacket.wValue  = SETUPDAT[2] + (SETUPDAT[3] << 8);
	SetupPacket.wIndex  = SETUPDAT[4] + (SETUPDAT[5] << 8);
	SetupPacket.wLength = SETUPDAT[6] + (SETUPDAT[7] << 8);

//	EP0BUF_ptr = EP0BUF;
	EP0BUF_count = SetupPacket.wLength;
	EP0BUF_len_requested = SetupPacket.wLength;
	EP0BUF_add_zero_packet = FALSE;
}

/*
 * Device request parser
 */
void ucore_SetupCommand(void)
{
	switch(SetupPacket.type) {
	case REQUEST_STANDARD:
			switch(SetupPacket.bRequest) {
			case USB_REQUEST_GET_STATUS:
rprintf("GET_STAT\r\n");
					if(!ucore_GetStatus()) {
						goto stall_i;
					}
					break;

			case USB_REQUEST_CLEAR_FEATURE:
rprintf("CLR_FEAT\r\n");
					if(!ucore_SetClrFeature(0)) {
						goto stall_i;
					}
					ucore_StatusInStage();
//					vcom_USB_Feature_Event();
					break;

			case USB_REQUEST_SET_FEATURE:
rprintf("SET_FEAT\r\n");
					if(!ucore_SetClrFeature(1)) {
						goto stall_i;
					}

					ucore_StatusInStage();
//					vcom_USB_Feature_Event();
					break;

			case USB_REQUEST_SET_ADDRESS:
rprintf("SET_ADDR\r\n");
					switch(SetupPacket.recip) {
					case REQUEST_TO_DEVICE:
							USB_DeviceAddress = 0x80 | (SetupPacket.wValue & 0xff);
							ucore_StatusInStage();
							break;
					default:
							goto stall_i;
					}
					break;

			case USB_REQUEST_GET_DESCRIPTOR:
rprintf("GET_DESC\r\n");
					if(!ucore_GetDescriptor()) {
						goto stall_i;
					}
					break;

			case USB_REQUEST_SET_DESCRIPTOR:
rprintf("SET_DESC\r\n");
///*stall_o:*/		uhw_SetStallEP(EP_CTRL_OUT);
					ucore_STALL_EP0();
					EP0BUF_count = 0;
					break;

			case USB_REQUEST_GET_CONFIGURATION:
rprintf("GET_CONF\r\n");
					switch(SetupPacket.recip) {
					case REQUEST_TO_DEVICE:
							EP0BUF_ptr = &USB_Configuration;
finish_to_parse_setup_request();
							ucore_DataInStage();
							break;
					default:
							goto stall_i;
					}
					break;

			case USB_REQUEST_SET_CONFIGURATION:
rprintf("SET_CONF\r\n");
					switch(SetupPacket.recip) {
					case REQUEST_TO_DEVICE:
							if(!ucore_SetConfiguration()) {
								goto stall_i;
							}
							ucore_StatusInStage();
//							vcom_USB_Configure_Event();
							break;
					default:
							goto stall_i;
					}
					break;

			case USB_REQUEST_GET_INTERFACE:
rprintf("GET_INTF\r\n");
					switch(SetupPacket.recip) {
					case REQUEST_TO_INTERFACE:
							if(
								(USB_Configuration != 0) &&
								((SetupPacket.wIndex & 0xff) < USB_NumInterfaces)
							) {
								EP0BUF_ptr = USB_AltSetting + (SetupPacket.wIndex & 0xff);
finish_to_parse_setup_request();
								ucore_DataInStage();
							} else {
								goto stall_i;
							}
							break;
					default:
							goto stall_i;
					}
					break;

			case USB_REQUEST_SET_INTERFACE:
rprintf("SET_INTF\r\n");
					switch(SetupPacket.recip) {
					case REQUEST_TO_INTERFACE:
							if(!ucore_SetInterface()) {
								goto stall_i;
							}
							ucore_StatusInStage();
//							vcom_USB_Interface_Event();
							break;
					default:
							goto stall_i;
					}
					break;

			default:
					goto stall_i;
			}
			break;

	case REQUEST_CLASS:
#if UDR_NEED_CLASS_SUPPORT
			switch(SetupPacket.recip) {
			case REQUEST_TO_INTERFACE:
					//break;
			default:
					goto stall_i;
			}
#else /* UDR_NEED_CLASS_SUPPORT */
			goto stall_i;
#endif /* UDR_NEED_CLASS_SUPPORT */

	case REQUEST_VENDOR:
#if UDR_NEED_VENDOR_SUPPORT
			UARTSettingCmd = SetupPacket.bRequest & 0xFF;
			UARTSettingData = SetupPacket.wValue & 0xFF;
			if(!vcom_SetSIOSetup( UARTSettingCmd, UARTSettingData )) {
				goto stall_i;
			}
finish_to_parse_setup_request();
			ucore_DataInStage();
#endif /* UDR_NEED_VENDOR_SUPPORT */
			break;

	default:
stall_i:
			ucore_STALL_EP0();
//			uhw_SetStallEP(0x80);
			EP0BUF_count = 0;
			break;
	}
rprintf("@\r\n");
}

/*
 *  Get Status USB Request
 *    Parameters:      None (global SetupPacket)
 *    Return Value:    TRUE - Success, FALSE - Error
 */
static int ucore_GetStatus(void)
{
	u_long n, m;

	switch(SetupPacket.recip) {
	case REQUEST_TO_DEVICE:
			EP0BUF_ptr = (u_char *)&USB_DeviceStatus;
finish_to_parse_setup_request();
			ucore_DataInStage();
			break;

	case REQUEST_TO_INTERFACE:
			if(
				(USB_Configuration != 0) &&
				((SetupPacket.wIndex & 0xff) < USB_NumInterfaces)
			) {
				EP0BUF[0] = 0;
				EP0BUF[1] = 0;
				EP0BUF_ptr = EP0BUF;
finish_to_parse_setup_request();
				ucore_DataInStage();
			} else {
				return(FALSE);
			}
			break;

	case REQUEST_TO_ENDPOINT:
			n = SetupPacket.wIndex & 0x8F;
			m = (n & 0x80) ? ((1 << 16) << n) : (1 << n);
			if(
				((USB_Configuration != 0) || ((n & 0x0F) == 0)) &&
				(USB_EndPointMask & m)
			) {
				EP0BUF[0] = (USB_EndPointHalt & m) ? 1 : 0;
				EP0BUF[1] = 0;
				EP0BUF_ptr = EP0BUF;
finish_to_parse_setup_request();
				ucore_DataInStage();
			} else {
				return(FALSE);
			}
			break;

	default:
			return(FALSE);
	}

	return(TRUE);
}

/*
 *  Set/Clear Feature USB Request
 *    Parameters:      sc:    0 - Clear, 1 - Set
 *                            None (global SetupPacket)
 *    Return Value:    TRUE - Success, FALSE - Error
 */
static int ucore_SetClrFeature(u_long sc)
{
	u_long n, m;

	switch(SetupPacket.recip) {
	case REQUEST_TO_DEVICE:
			if(SetupPacket.wValue == USB_FEATURE_REMOTE_WAKEUP) {
				if(sc) {
					USB_DeviceStatus |=  USB_GETSTATUS_REMOTE_WAKEUP;
				} else {
					USB_DeviceStatus &= ~USB_GETSTATUS_REMOTE_WAKEUP;
				}
			} else {
				return(FALSE);
			}
			break;

	case REQUEST_TO_INTERFACE:
			return(FALSE);

	case REQUEST_TO_ENDPOINT:
			n = SetupPacket.wIndex & 0x8F;
			m = (n & 0x80) ? ((1 << 16) << n) : (1 << n);
			if(
				(USB_Configuration != 0) &&
				((n & 0x0F) != 0) &&
				(USB_EndPointMask & m)
			) {
				if(SetupPacket.wValue == USB_FEATURE_ENDPOINT_STALL) {
					if(sc) {
						uhw_SetStallEP(n);
						USB_EndPointHalt |=  m;
					} else {
						uhw_ClrStallEP(n);
						USB_EndPointHalt &= ~m;
					}
				} else {
					return(FALSE);
				}
			} else {
				return(FALSE);
			}
			break;

	default:
			return(FALSE);
	}
	return(TRUE);
}

/*
 *  Get Descriptor USB Request
 *    Parameters:      None (global SetupPacket)
 *    Return Value:    TRUE - Success, FALSE - Error
 */
static int ucore_GetDescriptor(void)
{
	u_char *pD;
	u_long len, n;
	u_short total_len;

	switch(SetupPacket.recip) {
	case REQUEST_TO_DEVICE:
			switch(SetupPacket.wValue >> 8) {
			case USB_DEVICE_DESCRIPTOR_TYPE:
rprintf(".device\r\n");
					EP0BUF_ptr = (u_char *)USB_DeviceDescriptor;
					len = USB_DEVICE_DESC_SIZE;
					break;

			case USB_CONFIGURATION_DESCRIPTOR_TYPE:
rprintf(".config\r\n");
					pD = (u_char *)USB_ConfigDescriptor;
					for(n=0; n != (SetupPacket.wValue & 0xff); n++) {
						if(pD[/*bLength*/ 0] != 0) {
							total_len = (pD[2]<<0) + (pD[3]<<8);	/* wTotalLength */
							pD += total_len;
						}
					}
					if(pD[/*bLength*/ 0] == 0) {
						return(FALSE);
					}
					EP0BUF_ptr = pD;
					len = (pD[2]<<0) + (pD[3]<<8);	/* wTotalLength */
					break;

			case USB_STRING_DESCRIPTOR_TYPE:
rprintf(".string\r\n");
//					EP0BUF_ptr = (u_char *)USB_StringDescriptor + (SetupPacket.wValue & 0xff);
					EP0BUF_ptr = ucore_GetStringDscr(USB_StringDescriptor, (SetupPacket.wValue & 0xff));
					len = EP0BUF_ptr[0];		/* bLength */
					break;

			default:
					return(FALSE);
			}
			break;

	case REQUEST_TO_INTERFACE:
			switch(SetupPacket.wValue >> 8) {
			default:
					return(FALSE);
			}
			break;

	default:
			return(FALSE);
	}

	if(EP0BUF_count > len) {
		EP0BUF_count = len;
	}
finish_to_parse_setup_request();
	ucore_DataInStage();

	return(TRUE);
}

/*
 * search Nth string descriptor
 */
static u_char *ucore_GetStringDscr(u_char* top_string_desc, u_char index)
{
	u_char *dscr;

	dscr = top_string_desc;

	while(dscr[1 /*type*/] == USB_STRING_DESCRIPTOR_TYPE) {
		if(index == 0)
			return(dscr);
		index--;
		dscr += dscr[0 /*length*/];
	}

	return(NULL);
}
//STRINGDSCR *EZUSB_GetStringDscr(BYTE StrIdx)
//{
//	STRINGDSCR xdata *dscr;
//
//	dscr = (STRINGDSCR xdata *)pStringDscr;
//
//	while(dscr->type == STRING_DSCR) {
//		if(StrIdx == 0)
//			return(dscr);
//		StrIdx--;
//		dscr = (STRINGDSCR *)((WORD)dscr + dscr->length);
//	}
//
//	return(NULL);
//}

/*
 *  Set Configuration USB Request
 *    Parameters:      None (global SetupPacket)
 *    Return Value:    TRUE - Success, FALSE - Error
 */
static int ucore_SetConfiguration(void)
{
	u_char *pD;
	u_long alt, n, m;
	u_short total_len;

	if(SetupPacket.wValue & 0xff) {
		pD = (u_char *)USB_ConfigDescriptor;
		while(pD[/*bLength*/ 0]) {
			switch(pD[/*bDescriptorType*/ 1]) {
			case USB_CONFIGURATION_DESCRIPTOR_TYPE:
					if(pD[/*bConfigurationValue*/ 5] == (SetupPacket.wValue & 0xff)) {
						USB_Configuration = SetupPacket.wValue & 0xff;
						USB_NumInterfaces = pD[/*bNumInterfaces*/ 4];
						for(n=0; n<USB_N_IF; n++) {
							USB_AltSetting[n] = 0;
						}
						for(n=1; n<16; n++) {
							if(USB_EndPointMask & (1 << n)) {
								uhw_DisableEP(n);
							}
							if(USB_EndPointMask & ((1 << 16) << n)) {
								uhw_DisableEP(n | 0x80);
							}
						}
						USB_EndPointMask = 0x00010001;
						USB_EndPointHalt = 0x00000000;
						uhw_Configure(TRUE);

						if(pD[/*bmAttributes*/ 7] & USB_CONFIG_SELF_POWERED) {
							USB_DeviceStatus |=  USB_GETSTATUS_SELF_POWERED;
						} else {
							USB_DeviceStatus &= ~USB_GETSTATUS_SELF_POWERED;
						}
					} else {
						total_len = (pD[2]<<0) + (pD[3]<<8);	/* wTotalLength */
						pD += total_len;
						continue;
					}
					break;

			case USB_INTERFACE_DESCRIPTOR_TYPE:
					alt = pD[/*bAlternateSetting*/ 3];
					break;

			case USB_ENDPOINT_DESCRIPTOR_TYPE:
					if(alt == 0) {
						n = pD[/*bEndpointAddress*/ 2] & 0x8F;
						m = (n & 0x80) ? ((1 << 16) << n) : (1 << n);
						USB_EndPointMask |= m;
						uhw_ConfigEP(pD);
						uhw_EnableEP(n);
						uhw_ResetEP(n);
					}
					break;
			}

			pD += pD[/*bLength*/ 0];
		}
	} else {
		USB_Configuration = 0;
		for(n=1; n<16; n++) {
			if(USB_EndPointMask & (1 << n)) {
				uhw_DisableEP(n);
			}
			if(USB_EndPointMask & ((1 << 16) << n)) {
				uhw_DisableEP(n | 0x80);
			}
		}
		USB_EndPointMask  = 0x00010001;
		USB_EndPointHalt  = 0x00000000;
		uhw_Configure(FALSE);
	}

	if(USB_Configuration == (SetupPacket.wValue & 0xff)) {
		return(TRUE);
	} else {
		return(FALSE);
	}
}

/*
 *  Set Interface USB Request
 *    Parameters:      None (global SetupPacket)
 *    Return Value:    TRUE - Success, FALSE - Error
 */
static int ucore_SetInterface(void)
{
	u_char *pD;
	u_long ifn, alt, old, msk, n, m;
	int set;
	u_short total_len;

	if(USB_Configuration == 0)
		return(FALSE);

	set = FALSE;
	pD = (u_char *)USB_ConfigDescriptor;

	while(pD[/*bLength*/ 0]) {
		switch(pD[/*bDescriptorType*/ 1]) {
		case USB_CONFIGURATION_DESCRIPTOR_TYPE:
				if(pD[/*bConfigurationValue*/ 5] != USB_Configuration) {
					total_len = (pD[2]<<0) + (pD[3]<<8);	/* wTotalLength */
					pD += total_len;
					continue;
				}
				break;

		case USB_INTERFACE_DESCRIPTOR_TYPE:
				ifn = pD[/*bInterfaceNumber*/ 2];
				alt = pD[/*bAlternateSetting*/ 3];
				msk = 0;
				if(
					(ifn == (SetupPacket.wIndex & 0xff)) &&
					(alt == (SetupPacket.wValue & 0xff))
				) {
					set = TRUE;
					old = USB_AltSetting[ifn];
					USB_AltSetting[ifn] = (u_char)alt;
				}
				break;

		case USB_ENDPOINT_DESCRIPTOR_TYPE:
				if(ifn == (SetupPacket.wIndex & 0xff)) {
					n = pD[/*bEndpointAddress*/ 2] & 0x8F;
					m = (n & 0x80) ? ((1 << 16) << n) : (1 << n);
					if(alt == (SetupPacket.wValue & 0xff)) {
						USB_EndPointMask |=  m;
						USB_EndPointHalt &= ~m;
						uhw_ConfigEP(pD);
						uhw_EnableEP(n);
						uhw_ResetEP(n);
						msk |= m;
					} else if((alt == old) && ((msk & m) == 0)) {
						USB_EndPointMask &= ~m;
						USB_EndPointHalt &= ~m;
						uhw_DisableEP(n);
					}
				}
				break;
		}
		pD += pD[/*bLength*/ 0];
	}
	return(set);
}

/*
 * stall both End Point 0 in/out
 */
void ucore_STALL_EP0(void)
{
	uhw_SetStallEP(EP_CTRL_IN);		/* for IN */
	uhw_SetStallEP(EP_CTRL_OUT);		/* for OUT */
rprintf("STALL_EP0\r\n");
		/*
		 * the stall bits are automatically cleared when next SETUP received
		 */
}
